package drop;
import java.awt.Color;
import java.awt.Dimension;
import java.lang.*;
import java.util.Arrays;
import figure.Point3D;
import figure.Polygon3D;
import figure.Rotate;
import figure.Vector3D;
import math.InterPolation;
import shading.FlatShading;
import shading.GouraudShading;
import shading.Shading;

//--------------------------------------------------------------------//
/**
 * tHf[^fNX
 */
public class DropModel{

/** tbgVF[fBO\ int ^ */
  public static final int FLAT=1;

/** O[[VF[fBO\ int ^ */
  public static final int GOURAUD=2;

/** tH`̃f[^ */
  public static final int NMAX=181;

/** tH */
  protected Rotate drop;

/**  */
  protected Polygon3D floor;

/** _ */
  protected Point3D viewPoint;

/**  */
  protected Point3D lightPoint;

/** ʃTCY */
  protected Dimension size;

/** ʃTCY */
  protected double frameSize;

/** tH̍ */
  protected double height;

/** tH̑̐ */
  protected double volume;

/** ̐F */
  protected Color floorColor=new Color(51,51,255);

/** tH̐F */
  protected Color dropColor=Color.CYAN;

/**
 * tHf[^f\z
 * @param size ʃTCY
 */
  public DropModel(Dimension size){
    this.size=size;
    frameSize=(double)Math.min(size.width,size.height);
    height=0.0;
    volume=0.0;
    setArround();
  }

//--------------------------------------------------------------------//
/**
 * ӂݒ肷
 */
  protected void setArround(){
// ݒ肷
    double floorSize=0.8*frameSize;
    Point3D p[]={new Point3D(floorSize,floorSize,0.0),
      	      	 new Point3D(-floorSize,floorSize,0.0),
      	      	 new Point3D(-floorSize,-floorSize,0.0),
      	      	 new Point3D(floorSize,-floorSize,0.0)};
    floor=new Polygon3D(4,p);
    floor.setColor(floorColor);

// _ݒ肷
    double viewCoord=1.6*frameSize;
    viewPoint=new Point3D(viewCoord,viewCoord,viewCoord);

// ݒ肷
    double lightx=0.0;
    double lighty=1.4*frameSize;
    double lightz=3.0*frameSize;
    lightPoint=new Point3D(lightx,lighty,lightz);
  }

//--------------------------------------------------------------------//
/**
 * tHݒ肷
 * @param beta W 
 * @param gamma0 ڐGp
 */
  public void setDrop(double beta,double gamma0){
    double r[]=new double[NMAX];
    double f[]=new double[NMAX];
    double df[]=new double[NMAX];
    double gamma[]=new double[NMAX];
    double volume[]=new double[NMAX];

    double dth=Math.PI/(double)(NMAX-1);
    int n=DropShape.ABMMethod(NMAX,dth,r,f,df,gamma,gamma0,beta);
    double rbc=f[n-1];
    for(int i=0;i<n;i++){
      double th=r[i];
      double rr=f[i];
      r[i]=rr*Math.sin(th);
      f[i]=rr*Math.cos(th);
    }
    double dv=Math.PI/3.0;
    volume[0]=0.0;
    for(int i=1;i<n;i++){
      volume[i]=volume[i-1]+dv*(f[i-1]-f[i])*(r[i-1]*(r[i-1]+r[i])
      	      	      	   +r[i]*r[i]);
    }
    double r0=0.0;
    double f0=0.0;
    if((gamma0<gamma[n-1]) && (rbc>Double.MIN_VALUE)){
      r0=InterPolation.linear(n,gamma0,gamma,r);
      f0=InterPolation.linear(n,gamma0,gamma,f);
      this.volume=InterPolation.linear(n,gamma0,gamma,volume)/(r0*r0*r0);
    }
    else{
      r0=r[n-2];
      f0=f[n-2];
      this.volume=volume[n-2]/(r0*r0*r0);
    }
    height=Math.abs((f0-1.0)/r0);
    double rr=0.3*frameSize/r0;
    int ii=1;
    while(f0<f[ii]) ii++;
    n=ii+1;
    r[ii]=r0;
    f[ii]=f0;
    for(int i=0;i<n;i++){
      r[i]*=rr;
      f[i]=rr*(f[i]-f0);
    }
    Point3D org=new Point3D();
    Vector3D axis=new Vector3D(0.0,0.0,1.0);
    drop=new Rotate(org,axis,n,f,r);
    drop.setColor(dropColor);
  }

//--------------------------------------------------------------------//
/**
 * _ʒu]
 * @param theta ox
 * @param phy ܓx
 */
  public void viewRotate(double theta,double phy){
    double dxy=viewPoint.x*viewPoint.x+viewPoint.y*viewPoint.y;
    if(dxy>Double.MIN_VALUE){
      dxy=Math.sqrt(dxy);
      double th1=Math.acos(viewPoint.x/dxy);
      if(viewPoint.y<0.0) th1=-th1;
      viewPoint.rotZ(-th1);
      double th2=Math.toDegrees(Math.atan(viewPoint.z/dxy));
      phy=Math.max(Math.min(phy,89.0+th2),-89.0+th2);
      viewPoint.rotY(Math.toRadians(phy));
      viewPoint.rotZ(th1);
    }
    viewPoint.rotZ(Math.toRadians(theta));
  }

//--------------------------------------------------------------------//
/**
 * VF[fBÕsNZf[^Ԃ
 * @param type VF[fBO̎ނ\ int ^
 * @param perspect ϊsꍇ true
 * @param bgColor wiF
 * @return VF[fBÕsNZf[^
 */
  public int[] getPixel(int type,boolean perspect,Color bgColor){
    int pixel[]=new int[size.width*size.height];
    double zbuf[]=new double[pixel.length];
    Arrays.fill(pixel,bgColor.getRGB());
    Arrays.fill(zbuf,-Double.MAX_VALUE);
    Shading shading=null;
    switch(type){
      case FLAT:
      	shading=new FlatShading(size.width,size.height,viewPoint,
      	      	      	      	lightPoint,perspect);
      	break;
      case GOURAUD:
      	shading=new GouraudShading(size.width,size.height,viewPoint,
      	      	      	      	   lightPoint,perspect);
      	break;
    }
    shading.shading(drop,pixel,zbuf);
    shading.shading(floor,pixel,zbuf);
    return pixel;
  }
}
